@@ -18,6 +18,7 @@ from nbxmpp.task import Task
 
 from gajim.common import app
 from gajim.common import events
+from gajim.common import ged
 from gajim.common.commands import ChatCommands
 from gajim.common.const import CallType
 from gajim.common.ged import EventHelper
@@ -38,11 +39,11 @@ from gajim.common.util.preview import split_geo_uri
 from gajim.common.util.text import remove_invalid_xml_chars
 
 from gajim.gtk.activity_page import ActivityPage
+from gajim.gtk.alert import InformationAlertDialog
 from gajim.gtk.chat_banner import ChatBanner
 from gajim.gtk.chat_function_page import ChatFunctionPage
 from gajim.gtk.chat_function_page import FunctionMode
 from gajim.gtk.control import ChatControl
-from gajim.gtk.dialogs import SimpleDialog
 from gajim.gtk.message_actions_box import MessageActionsBox
 from gajim.gtk.message_input import MessageInputTextView
 from gajim.gtk.util.classes import SignalManager
@@ -53,6 +54,8 @@ log = logging.getLogger("gajim.gtk.chatstack")
 
 
 class ChatStack(Gtk.Stack, EventHelper, SignalManager):
+    __gtype_name__ = "ChatStack"
+
     def __init__(self):
         Gtk.Stack.__init__(self, hexpand=True, vexpand=True)
         EventHelper.__init__(self)
@@ -94,13 +97,13 @@ class ChatStack(Gtk.Stack, EventHelper, SignalManager):
         dnd_label.set_max_width_chars(40)
         dnd_label.set_vexpand(True)
         dnd_label.set_valign(Gtk.Align.START)
-        dnd_label.add_css_class("bold16")
+        dnd_label.add_css_class("title-3")
 
         self._drop_area = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=18)
         self._drop_area.set_visible(False)
         self._drop_area.set_hexpand(True)
         self._drop_area.set_vexpand(True)
-        self._drop_area.add_css_class("solid-background")
+        self._drop_area.add_css_class("background")
         self._drop_area.append(dnd_icon)
         self._drop_area.append(dnd_label)
 
@@ -118,14 +121,13 @@ class ChatStack(Gtk.Stack, EventHelper, SignalManager):
 
         self.add_named(overlay, "controls")
 
-        self._connect_actions()
-
         self.register_events(
             [
                 ("message-received", 85, self._on_message_received),
                 ("muc-disco-update", 85, self._on_muc_disco_update),
                 ("account-connected", 85, self._on_account_state),
                 ("account-disconnected", 85, self._on_account_state),
+                ("register-actions", ged.GUI1, self._on_register_actions),
             ]
         )
 
@@ -135,6 +137,33 @@ class ChatStack(Gtk.Stack, EventHelper, SignalManager):
         self.unregister_events()
         app.check_finalize(self)
 
+    def _on_register_actions(self, _event: events.RegisterActions) -> None:
+        actions = [
+            "send-file",
+            "send-file-httpupload",
+            # 'send-file-jingle',
+            "start-video-call",
+            "start-voice-call",
+            "send-message",
+            "show-contact-info",
+            "muc-change-nickname",
+            "muc-invite",
+            "muc-contact-info",
+            "muc-execute-command",
+            "muc-ban",
+            "muc-kick",
+            "muc-change-role",
+            "muc-change-affiliation",
+            "muc-request-voice",
+            "quote-next",
+            "quote-prev",
+        ]
+
+        for action in actions:
+            action = app.window.lookup_action(action)
+            assert action is not None
+            action.connect("activate", self._on_action)
+
     def _get_current_contact(self) -> ChatContactT:
         assert self._current_contact is not None
         return self._current_contact
@@ -249,8 +278,14 @@ class ChatStack(Gtk.Stack, EventHelper, SignalManager):
         clipboard: Gdk.Clipboard,
         result: Gio.AsyncResult,
     ) -> None:
-        text = clipboard.read_text_finish(result)
+        try:
+            text = clipboard.read_text_finish(result)
+        except Exception as e:
+            InformationAlertDialog(_("Pasting Content Failed"), _("Error: %s") % e)
+            return
+
         if text is None:
+            log.info("No text pasted")
             return
 
         # Reset primary clipboard to what it was before switching chats,
@@ -382,7 +417,8 @@ class ChatStack(Gtk.Stack, EventHelper, SignalManager):
             )
 
             if not mds_assist_sent and stanza_id is not None:
-                client.get_module("MDS").set_mds(contact.jid, stanza_id)
+                by = contact.jid if isinstance(contact, GroupchatContact) else None
+                client.get_module("MDS").set_mds(contact.jid, stanza_id, by)
 
             return
 
@@ -486,33 +522,6 @@ class ChatStack(Gtk.Stack, EventHelper, SignalManager):
             )
         )
 
-    def _connect_actions(self) -> None:
-        actions = [
-            "send-file",
-            "send-file-httpupload",
-            # 'send-file-jingle',
-            "start-video-call",
-            "start-voice-call",
-            "send-message",
-            "show-contact-info",
-            "muc-change-nickname",
-            "muc-invite",
-            "muc-contact-info",
-            "muc-execute-command",
-            "muc-ban",
-            "muc-kick",
-            "muc-change-role",
-            "muc-change-affiliation",
-            "muc-request-voice",
-            "quote-next",
-            "quote-prev",
-        ]
-
-        for action in actions:
-            action = app.window.lookup_action(action)
-            assert action is not None
-            action.connect("activate", self._on_action)
-
     def _update_base_actions(self, contact: ChatContactT) -> None:
         client = app.get_client(contact.account)
         online = app.account_is_connected(contact.account)
@@ -533,6 +542,7 @@ class ChatStack(Gtk.Stack, EventHelper, SignalManager):
         )
 
         app.window.get_action("correct-message").set_enabled(online)
+        app.window.get_action("retract-message").set_enabled(online)
 
     def _update_chat_actions(self, contact: BareContact) -> None:
         online = app.account_is_connected(contact.account)
@@ -712,7 +722,7 @@ class ChatStack(Gtk.Stack, EventHelper, SignalManager):
                         group=muc.name,
                     ),
                 )
-                SimpleDialog(_("Error"), f"{error_message}\n{error}")
+                InformationAlertDialog(_("Error"), f"{error_message}\n{error}")
         else:
             log.debug("Affiliation/role change success: %s", result)
 
@@ -803,7 +813,7 @@ class ChatStack(Gtk.Stack, EventHelper, SignalManager):
 
         elif encryption:
             if encryption not in app.plugin_manager.encryption_plugins:
-                SimpleDialog(
+                InformationAlertDialog(
                     _("Encryption Error"), _("Missing necessary encryption plugin")
                 )
                 return
