@@ -6,10 +6,12 @@ from __future__ import annotations
 
 import logging
 import textwrap
+from datetime import datetime
 from datetime import timedelta
 
 from gi.repository import GLib
 from gi.repository import Gtk
+from nbxmpp.namespaces import Namespace
 
 from gajim.common import app
 from gajim.common.const import AvatarSize
@@ -26,7 +28,9 @@ from gajim.common.storage.archive.const import MessageType
 from gajim.common.storage.archive.models import Message
 from gajim.common.types import ChatContactT
 from gajim.common.util.muc import message_needs_highlight
+from gajim.common.util.text import format_fingerprint
 from gajim.common.util.user_strings import get_moderation_text
+from gajim.common.util.user_strings import get_retraction_text
 
 from gajim.gtk.conversation.message_widget import MessageWidget
 from gajim.gtk.conversation.reactions_bar import ReactionsBar
@@ -36,44 +40,47 @@ from gajim.gtk.conversation.rows.widgets import DateTimeLabel
 from gajim.gtk.conversation.rows.widgets import MessageIcons
 from gajim.gtk.conversation.rows.widgets import MessageRowActions
 from gajim.gtk.conversation.rows.widgets import NicknameLabel
+from gajim.gtk.menus import GajimMenu
 from gajim.gtk.menus import get_chat_row_menu
 from gajim.gtk.preview import PreviewWidget
+from gajim.gtk.referenced_message import ReferencedMessageNotFoundWidget
 from gajim.gtk.referenced_message import ReferencedMessageWidget
-from gajim.gtk.util import format_fingerprint
-from gajim.gtk.util import GajimPopover
-from gajim.gtk.util import get_avatar_for_message
-from gajim.gtk.util import get_contact_name_for_message
+from gajim.gtk.util.misc import container_remove_all
+from gajim.gtk.util.misc import get_avatar_for_message
+from gajim.gtk.util.misc import get_contact_name_for_message
+from gajim.gtk.widgets import GajimPopover
 
-log = logging.getLogger('gajim.gtk.conversation.rows.message')
+log = logging.getLogger("gajim.gtk.conversation.rows.message")
 
 MERGE_TIMEFRAME = timedelta(seconds=120)
 
 
 class MessageRow(BaseRow):
-    def __init__(self,
-                 contact: ChatContactT,
-                 message: Message
-                 ) -> None:
+    def __init__(self, contact: ChatContactT, message: Message) -> None:
 
         BaseRow.__init__(self, contact.account)
         self.set_selectable(True)
-        self.type = 'chat'
+        self.type = "chat"
         self._contact = contact
         self._message = message
 
         self.timestamp = message.timestamp.astimezone()
         self.db_timestamp = message.timestamp.timestamp()
-        self.stanza_id = message.stanza_id
         self.direction = ChatDirection(message.direction)
         self._is_outgoing = self.direction == ChatDirection.OUTGOING
 
         self.orig_pk = message.pk
+        self.pk = self.orig_pk
+
+        self.encryption = None
+        self.securitylabel = None
 
         assert message.text is not None
         self._original_text = message.text
         self._original_message = message
 
-        self._is_moderated = message.moderation is not None
+        self._is_retracted = bool(message.moderation or message.retraction)
+        self._is_blocked = False
         self._has_receipt = False
 
         self._avatar_box = AvatarBox(contact)
@@ -92,25 +99,18 @@ class MessageRow(BaseRow):
         self._reactions_bar = ReactionsBar(self, self._contact)
         self.grid.attach(self._reactions_bar, 1, 2, 1, 1)
 
-        self._set_content(message)
+        self._redraw_content()
+
+    def do_unroot(self) -> None:
+        BaseRow.do_unroot(self)
 
     @classmethod
-    def from_db_row(cls,
-        contact: ChatContactT,
-        message: Message
-    ) -> MessageRow:
-
+    def from_db_row(cls, contact: ChatContactT, message: Message) -> MessageRow:
         return cls(contact, message)
 
     @property
-    def last_message_id(self) -> str | None:
-        if self._corr_message is None:
-            return self._original_message.id
-        return self._corr_message.id
-
-    @property
-    def message_id(self) -> str | None:
-        return self._original_message.id
+    def state(self) -> int:
+        return self._message.state
 
     @property
     def has_receipt(self) -> bool:
@@ -120,32 +120,50 @@ class MessageRow(BaseRow):
     def is_merged(self) -> bool:
         return self._merged
 
-    def refresh(self, *, complete: bool = True) -> None:
-        original_message = app.storage.archive.get_message_with_pk(
-            self.orig_pk)
-        assert original_message is not None
-        self._original_message = original_message
-        if complete:
-            self._set_content(original_message)
+    @property
+    def occupant_id(self) -> str | None:
+        if self._message.occupant is None:
+            return None
+        return self._message.occupant.id
 
-    def _set_content(self, message: Message) -> None:
+    def _redraw_content(self) -> None:
         self.set_merged(False)
-        self.get_style_context().remove_class('moderated-message')
-        self.get_style_context().remove_class('gajim-mention-highlight')
+        self.remove_css_class("retracted-message")
+        self.remove_css_class("gajim-mention-highlight")
 
-        for widget in self._meta_box.get_children():
-            widget.destroy()
+        container_remove_all(self._meta_box)
+        container_remove_all(self._bottom_box)
 
-        for widget in self._bottom_box.get_children():
-            widget.destroy()
+        message = self._original_message
 
-        self._corr_message = None
+        self.name = get_contact_name_for_message(message, self._contact)
+
+        avatar = get_avatar_for_message(
+            message, self._contact, self.get_scale_factor(), AvatarSize.ROSTER
+        )
+        self._avatar_box.set_from_paintable(avatar)
+        self._avatar_box.set_name(self.name)
+
+        self._meta_box.append(NicknameLabel(self.name, self._is_outgoing))
+        self._meta_box.append(DateTimeLabel(self.timestamp))
+
+        self._message_icons = MessageIcons()
+        self._meta_box.append(self._message_icons)
+
+        self._is_blocked = False
+        if message.occupant is not None and message.occupant.blocked:
+            self._set_blocked()
+            return
+
+        if message.is_retracted():
+            self._set_retracted(message)
+            return
 
         # From here on, if this is a correction all data must
         # be taken from the correction
-        if message.corrections:
-            message = message.get_last_correction()
-            self._corr_message = message
+        if corrected_message := message.get_last_correction():
+            message = corrected_message
+            self._message = corrected_message
 
         self.pk = message.pk
 
@@ -155,40 +173,20 @@ class MessageRow(BaseRow):
         assert message.text is not None
         self.text = message.text
 
-        self.name = get_contact_name_for_message(message, self._contact)
-
-        avatar = get_avatar_for_message(
-            message,
-            self._contact,
-            self.get_scale_factor(),
-            AvatarSize.ROSTER
-        )
-        self._avatar_box.set_from_surface(avatar)
-        self._avatar_box.set_name(self.name)
-
-        self._meta_box.pack_start(
-            NicknameLabel(
-                self.name,
-                self._is_outgoing
-            ), False, True, 0)
-        self._meta_box.pack_start(
-            DateTimeLabel(self.timestamp), False, True, 0)
-
-        self._message_icons = MessageIcons()
-        self._meta_box.pack_start(self._message_icons, False, True, 0)
-
         if app.preview_manager.is_previewable(self.text, message.oob):
             self._message_widget = PreviewWidget(self._contact.account)
             app.preview_manager.create_preview(
-                self.text,
-                self._message_widget,
-                self._is_outgoing,
-                self._muc_context)
+                self.text, self._message_widget, self._is_outgoing, self._muc_context
+            )
         else:
-            referenced_message = message.get_referenced_message()
-            if referenced_message is not None:
-                self._ref_message_widget = ReferencedMessageWidget(
-                    self._contact, referenced_message)
+            if message.reply is not None:
+                referenced_message = message.get_referenced_message()
+                if referenced_message is None:
+                    self._ref_message_widget = ReferencedMessageNotFoundWidget()
+                else:
+                    self._ref_message_widget = ReferencedMessageWidget(
+                        self._contact, referenced_message
+                    )
 
             self._message_widget = MessageWidget(self._contact.account)
             self._message_widget.add_with_styling(self.text, nickname=self.name)
@@ -197,24 +195,20 @@ class MessageRow(BaseRow):
 
         if self._ref_message_widget is not None:
             box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=3)
-            box.add(self._ref_message_widget)
-            box.add(self._message_widget)
-            self._bottom_box.add(box)
+            box.append(self._ref_message_widget)
+            box.append(self._message_widget)
+            self._bottom_box.append(box)
         else:
-            self._bottom_box.add(self._message_widget)
+            self._bottom_box.append(self._message_widget)
 
         self._set_text_direction(self.text)
 
-        if self._original_message.corrections:
+        if corrected_message:
             self._set_correction()
 
-        if message.moderation is not None:
-            self.set_moderated(get_moderation_text(
-                message.moderation.by, message.moderation.reason))
-
-        reactions = self._original_message.reactions
-        if reactions:
-            self._reactions_bar.update_from_reactions(reactions)
+        self._reactions_bar.update_from_reactions(
+            self._original_message.get_reactions()
+        )
 
         encryption_data = self._get_encryption_data(message.encryption)
         if encryption_data is not None:
@@ -226,22 +220,17 @@ class MessageRow(BaseRow):
             self._message_icons.set_security_label_data(*sec_label_data)
             self._message_icons.set_security_label_visible(True)
 
-        self.set_receipt(message.receipt is not None)
+        self._set_receipt(message.receipt is not None)
 
-        self.state = MessageState(message.state)
-
-        if (self._contact.is_groupchat and
-                self.direction == ChatDirection.OUTGOING):
-            self._message_icons.set_message_state_icon(self.state)
+        if self._contact.is_groupchat and self.direction == ChatDirection.OUTGOING:
+            self._message_icons.set_message_state_icon(MessageState(message.state))
 
         if message.error is not None:
             if message.error.text is not None:
-                error_text = f'{message.error.text} ({message.error.condition})'
+                error_text = f"{message.error.text} ({message.error.condition})"
             else:
                 error_text = message.error.condition
-            self.show_error(error_text)
-
-        self.show_all()
+            self._set_error(error_text)
 
     def _set_text_direction(self, text: str) -> None:
         if is_rtl_text(text):
@@ -253,41 +242,41 @@ class MessageRow(BaseRow):
 
     @property
     def _muc_context(self) -> str | None:
-        if isinstance(self._contact,
-                      GroupchatContact | GroupchatParticipant):
+        if isinstance(self._contact, GroupchatContact | GroupchatParticipant):
             return self._contact.muc_context
         return None
 
-    def show_chat_row_menu(
+    def get_chat_row_menu(
         self,
-        message_row_actions: MessageRowActions,
-        button: Gtk.Button
-    ) -> None:
-        menu = get_chat_row_menu(
+    ) -> GajimMenu:
+
+        copy_text = self._get_copy_text(self.get_text(), self.name, self.timestamp)
+
+        return get_chat_row_menu(
             contact=self._contact,
-            name=self.name,
-            text=self.get_text(),
-            timestamp=self.timestamp,
-            message_id=self._original_message.id,
-            stanza_id=self.stanza_id,
-            pk=self.orig_pk,
-            corrected_pk=self.pk,
-            state=self.state,
-            is_moderated=self._is_moderated,
+            copy_text=copy_text,
+            message=self._message,
+            original_message=self._original_message,
         )
 
-        popover = GajimPopover(menu, relative_to=button, position=Gtk.PositionType.TOP)
-        popover.connect(
-            'closed',
-            self._on_more_menu_popover_closed,
-            message_row_actions
-        )
-        popover.popup()
+    @staticmethod
+    def _get_copy_text(text: str, name: str, timestamp: datetime) -> str | None:
+        # Text can be an empty string
+        # e.g. if a preview has not been loaded yet
+        if not text:
+            return None
+
+        timestamp_formatted = timestamp.strftime(app.settings.get("date_time_format"))
+
+        copy_text = f"{timestamp_formatted} - {name}: "
+        if text.startswith(("```", "> ")):
+            # Prepend a line break in order to keep code block/quotes rendering
+            copy_text += "\n"
+        copy_text += text
+        return copy_text
 
     def _on_more_menu_popover_closed(
-        self,
-        _popover: GajimPopover,
-        message_row_actions: MessageRowActions
+        self, _popover: GajimPopover, message_row_actions: MessageRowActions
     ) -> None:
         message_row_actions.hide_actions()
 
@@ -300,15 +289,15 @@ class MessageRow(BaseRow):
             self._message_widget.set_selectable(True)
 
     def _get_security_labels_data(
-            self,
-            security_labels: mod.SecurityLabel | None
-            ) -> tuple[str, str] | None:
+        self, security_labels: mod.SecurityLabel | None
+    ) -> tuple[str, str] | None:
 
         if security_labels is None:
             return None
 
-        if not app.settings.get_account_setting(self._account,
-                                                'enable_security_labels'):
+        if not app.settings.get_account_setting(
+            self._account, "enable_security_labels"
+        ):
             return None
 
         displaymarking = GLib.markup_escape_text(security_labels.displaymarking)
@@ -317,7 +306,8 @@ class MessageRow(BaseRow):
             fgcolor = security_labels.fgcolor
             markup = (
                 f'<span size="small" bgcolor="{bgcolor}" '
-                f'fgcolor="{fgcolor}"><tt>{displaymarking}</tt></span>')
+                f'fgcolor="{fgcolor}"><tt>{displaymarking}</tt></span>'
+            )
             return displaymarking, markup
         return None
 
@@ -327,9 +317,9 @@ class MessageRow(BaseRow):
             return
 
         if message_needs_highlight(
-                text, self._contact.nickname, self._client.get_own_jid().bare):
-            self.get_style_context().add_class(
-                'gajim-mention-highlight')
+            text, self._contact.nickname, self._client.get_own_jid().bare
+        ):
+            self.add_css_class("gajim-mention-highlight")
 
     def is_same_sender(self, message: MessageRow) -> bool:
         return message.name == self.name
@@ -359,10 +349,10 @@ class MessageRow(BaseRow):
         return sec1.label_hash == sec2.label_hash
 
     def is_same_state(self, message: MessageRow) -> bool:
-        return message.state == self.state
+        return message.state == self._message.state
 
     def has_same_receipt_status(self, message: MessageRow) -> bool:
-        if not app.settings.get('positive_184_ack'):
+        if not app.settings.get("positive_184_ack"):
             return True
         return self._has_receipt == message.has_receipt
 
@@ -373,7 +363,7 @@ class MessageRow(BaseRow):
             return False
         if message.direction != self.direction:
             return False
-        if self._original_message.corrections:
+        if self._message.correction_id is not None:
             return False
         if not self.is_same_sender(message):
             return False
@@ -388,65 +378,112 @@ class MessageRow(BaseRow):
     def get_text(self) -> str:
         return self._message_widget.get_text()
 
-    def _get_encryption_data(self,
-                             encryption_data: mod.Encryption | None,
-                             ) -> tuple[str, str, str] | None:
+    def _get_encryption_data(
+        self,
+        encryption_data: mod.Encryption | None,
+    ) -> tuple[str, str, str] | None:
 
-        contact_encryption = self._contact.settings.get('encryption')
+        contact_encryption = self._contact.settings.get("encryption")
         if encryption_data is None:
             if not contact_encryption:
                 return None
 
-            icon = 'channel-insecure-symbolic'
-            color = 'unencrypted-color'
-            tooltip = _('Not encrypted')
+            icon = "channel-insecure-symbolic"
+            color = "unencrypted-color"
+            tooltip = _("Not encrypted")
         else:
-            tooltip = _('Encrypted (%s)') % (encryption_data.protocol)
-            icon, trust_tooltip, color = TRUST_SYMBOL_DATA[
-                Trust(encryption_data.trust)]
-            tooltip = f'{tooltip}\n{trust_tooltip}'
-            if encryption_data.key != 'Unknown':
+            tooltip = _("Encrypted (%s)") % (encryption_data.protocol)
+            icon, trust_tooltip, color = TRUST_SYMBOL_DATA[Trust(encryption_data.trust)]
+            tooltip = f"{tooltip}\n{trust_tooltip}"
+            if encryption_data.key != "Unknown":
                 fingerprint = format_fingerprint(encryption_data.key)
-                tooltip = f'{tooltip}\n<tt>{fingerprint}</tt>'
+                tooltip = f"{tooltip}\n<tt>{fingerprint}</tt>"
 
         return icon, color, tooltip
 
-    def set_acknowledged(self, stanza_id: str | None) -> None:
-        self.state = MessageState.ACKNOWLEDGED
-        self.stanza_id = stanza_id
-        self._message_icons.set_message_state_icon(self.state)
+    def set_acknowledged(self, pk: int) -> None:
+        if self._original_message.pk == pk:
+            message = self._original_message
+        elif self._message.pk == pk:
+            message = self._message
+        else:
+            raise ValueError("Acknowledged unknown message")
+
+        app.storage.archive.refresh(message, ["stanza_id", "state"])
+
+        if self._is_retracted:
+            return
+
+        self._message_icons.set_message_state_icon(MessageState(message.state))
+
+    def set_blocked(self, blocked: bool) -> None:
+        if blocked != self._is_blocked:
+            app.storage.archive.refresh(self._original_message, ["occupant"])
+            self._redraw_content()
+
+    def _set_receipt(self, value: bool) -> None:
+        if self._is_retracted:
+            return
 
-    def set_receipt(self, value: bool) -> None:
         self._has_receipt = value
         self._message_icons.set_receipt_icon_visible(value)
 
-    def show_error(self, tooltip: str) -> None:
+    def set_receipt(self, receipt_id: str) -> None:
+        if self._message.id == receipt_id:
+            self._set_receipt(True)
+
+    def _set_error(self, tooltip: str) -> None:
+        if self._is_retracted:
+            return
+
         self._message_icons.hide_message_state_icon()
         self._message_icons.set_error_icon_visible(True)
         self._message_icons.set_error_tooltip(tooltip)
 
+    def set_error(self, error_id: str, text: str) -> None:
+        if self._message.id == error_id:
+            self._set_error(text)
+            self.set_merged(False)
+
     def update_reactions(self) -> None:
-        self.refresh(complete=False)
-        self._reactions_bar.update_from_reactions(self._original_message.reactions)
+        app.storage.archive.refresh(
+            self._original_message, ["corrections", "reactions"]
+        )
+        if self._is_retracted:
+            return
+
+        self._reactions_bar.update_from_reactions(
+            self._original_message.get_reactions()
+        )
+
+    def update_retractions(self) -> None:
+        app.storage.archive.refresh(
+            self._original_message, ["corrections", "retraction", "moderation"]
+        )
+        self._redraw_content()
+
+    def update_corrections(self) -> None:
+        app.storage.archive.refresh(self._original_message, ["corrections"])
+        self._redraw_content()
 
     def send_reaction(self, emoji: str, toggle: bool = True) -> None:
-        '''Adds or removes 'emoji' from this message's reactions and sends the result.
+        """Adds or removes 'emoji' from this message's reactions and sends the result.
 
         Args:
           emoji: Reaction emoji to add or remove
           toggle: Whether an existing emoji should be removed from the set
-        '''
-        reaction_id = self.message_id
+        """
+        reaction_id = self._original_message.id
         if self._original_message.type == MessageType.GROUPCHAT:
-            reaction_id = self.stanza_id
+            reaction_id = self._original_message.stanza_id
 
         if reaction_id is None:
-            log.warning('No reaction id')
+            log.warning("No reaction id")
             return
 
         our_reactions = self._reactions_bar.get_our_reactions()
         if emoji in our_reactions and not toggle:
-            log.info('Not toggling reaction <%s>', emoji)
+            log.info("Not toggling reaction <%s>", emoji)
             return
 
         if emoji in our_reactions:
@@ -455,51 +492,108 @@ class MessageRow(BaseRow):
             our_reactions.add(emoji)
 
         client = app.get_client(self._contact.account)
-        client.get_module('Reactions').send_reaction(
+        client.get_module("Reactions").send_reaction(
             contact=self._contact, reaction_id=reaction_id, reactions=our_reactions
         )
 
-    def set_moderated(self, text: str) -> None:
+    def _set_retracted(self, message: Message) -> None:
+        if message.moderation is not None:
+            text = get_moderation_text(message.moderation.by, message.moderation.reason)
+
+        elif message.retraction is not None:
+            text = get_retraction_text(message.retraction.timestamp)
+
+        else:
+            raise ValueError("Unable to determine retraction text")
+
+        self._set_disabled(text)
+        self._is_retracted = True
+
+    def _set_blocked(self) -> None:
+        self._set_disabled(_("Messages from this users are blocked"))
+        self._is_blocked = True
+
+    def _set_disabled(self, text: str) -> None:
         self.text = text
 
-        if isinstance(self._message_widget, PreviewWidget):
-            self._message_widget.destroy()
-            self._message_widget = MessageWidget(self._account)
-            self._bottom_box.pack_start(self._message_widget, True, True, 0)
-            self._set_text_direction(text)
-
+        self._message_widget = MessageWidget(self._account)
         self._message_widget.add_with_styling(text)
-        self.get_style_context().add_class('moderated-message')
+        self._set_text_direction(text)
+        self._bottom_box.append(self._message_widget)
 
-        self._is_moderated = True
+        self._reactions_bar.set_visible(False)
+
+        self.add_css_class("retracted-message")
 
     def _set_correction(self) -> None:
-        original_text = textwrap.fill(self._original_text,
-                                      width=150,
-                                      max_lines=10,
-                                      placeholder='…')
+        original_text = textwrap.fill(
+            self._original_text, width=150, max_lines=10, placeholder="…"
+        )
         self._message_icons.set_correction_tooltip(
-            _('Message corrected. Original message:\n%s') % original_text)
+            _("Message corrected. Original message:\n%s") % original_text
+        )
         self._message_icons.set_correction_icon_visible(True)
 
     def update_avatar(self) -> None:
         avatar = get_avatar_for_message(
-            self._message,
+            self._original_message,
             self._contact,
             self.get_scale_factor(),
-            AvatarSize.ROSTER
+            AvatarSize.ROSTER,
         )
-        self._avatar_box.set_from_surface(avatar)
+        self._avatar_box.set_from_paintable(avatar)
 
     def set_merged(self, merged: bool) -> None:
         self._merged = merged
         if merged:
-            self.get_style_context().add_class('merged')
-            self._meta_box.set_no_show_all(True)
-            self._meta_box.hide()
+            self.add_css_class("merged")
+            self._meta_box.set_visible(False)
+            self._meta_box.set_visible(False)
         else:
-            self.get_style_context().remove_class('merged')
-            self._meta_box.set_no_show_all(False)
-            self._meta_box.show()
+            self.remove_css_class("merged")
+            self._meta_box.set_visible(True)
 
         self._avatar_box.set_merged(merged)
+
+    def can_reply(self) -> bool:
+        if self._is_retracted:
+            return False
+
+        if isinstance(self._contact, GroupchatContact):
+            if self._message.stanza_id is None:
+                return False
+
+            if not self._contact.is_joined:
+                return False
+
+            self_contact = self._contact.get_self()
+            assert self_contact is not None
+            return not self_contact.role.is_visitor
+
+        return self._original_message.id is not None
+
+    def can_react(self) -> bool:
+        if self._is_retracted:
+            return False
+
+        if not app.account_is_connected(self._contact.account):
+            return False
+
+        if isinstance(self._contact, GroupchatContact):
+            if not self._contact.is_joined:
+                return False
+
+            self_contact = self._contact.get_self()
+            assert self_contact is not None
+            if self_contact.role.is_visitor:
+                return False
+
+            if self._message.stanza_id is None:
+                return False
+
+            if self._contact.muc_context == "public":
+                return self._contact.supports(Namespace.OCCUPANT_ID)
+
+            return True
+
+        return self._original_message.id is not None
